@@ -41,6 +41,27 @@
 //!     .expire(key, 60).ignore()
 //!     .exec(&mut connection).unwrap();
 //! ```
+//!
+//! # Sending request to specific node
+//! In some cases you'd want to send a request to a specific node in the cluster, instead of
+//! letting the cluster connection decide by itself to which node it should send the request.
+//! This can happen, for example, if you want to send SCAN commands to each node in the cluster.
+//!
+//! ```rust,no_run
+//! use redis::Commands;
+//! use redis::cluster::ClusterClient;
+//! use redis::cluster_routing::{ RoutingInfo, SingleNodeRoutingInfo };
+//!
+//! let nodes = vec!["redis://127.0.0.1:6379/", "redis://127.0.0.1:6378/", "redis://127.0.0.1:6377/"];
+//! let client = ClusterClient::new(nodes).unwrap();
+//! let mut connection = client.get_connection().unwrap();
+//!
+//! let routing_info = RoutingInfo::SingleNode(SingleNodeRoutingInfo::ByAddress{
+//!     host: "redis://127.0.0.1".to_string(),
+//!     port: 6378
+//! });
+//! let _: redis::Value = connection.route_command(&redis::cmd("PING"), routing_info).unwrap();
+//! ```
 use std::cell::RefCell;
 use std::collections::HashSet;
 use std::str::FromStr;
@@ -64,15 +85,11 @@ use crate::{
     cluster_client::ClusterParams,
     cluster_routing::{Redirect, Route, RoutingInfo, SlotMap, SLOT_SIZE},
 };
-use rand::{seq::IteratorRandom, thread_rng, Rng};
+use rand::{rng, seq::IteratorRandom, Rng};
 
 pub use crate::cluster_client::{ClusterClient, ClusterClientBuilder};
 pub use crate::cluster_pipeline::{cluster_pipe, ClusterPipeline};
 
-#[cfg(feature = "tls-rustls")]
-use crate::tls::TlsConnParams;
-
-#[cfg(not(feature = "tls-rustls"))]
 use crate::connection::TlsConnParams;
 
 #[derive(Clone)]
@@ -84,7 +101,6 @@ enum Input<'a> {
     Cmd(&'a Cmd),
     Commands {
         cmd: &'a [u8],
-        route: SingleNodeRoutingInfo,
         offset: usize,
         count: usize,
     },
@@ -101,12 +117,7 @@ impl<'a> Input<'a> {
                 .req_command(cmd)
                 .and_then(|value| value.extract_error())
                 .map(Output::Single),
-            Input::Commands {
-                cmd,
-                route: _,
-                offset,
-                count,
-            } => connection
+            Input::Commands { cmd, offset, count } => connection
                 .req_packed_commands(cmd, *offset, *count)
                 .and_then(Value::extract_error_vec)
                 .map(Output::Multi),
@@ -114,7 +125,7 @@ impl<'a> Input<'a> {
     }
 }
 
-impl<'a> Routable for Input<'a> {
+impl Routable for Input<'_> {
     fn arg_idx(&self, idx: usize) -> Option<&[u8]> {
         match self {
             Input::Slice { cmd: _, routable } => routable.arg_idx(idx),
@@ -213,9 +224,79 @@ impl Connect for Connection {
     }
 }
 
-/// This represents a Redis Cluster connection. It stores the
-/// underlying connections maintained for each node in the cluster, as well
-/// as common parameters for connecting to nodes and executing commands.
+/// Options for creation of connection
+#[derive(Clone, Default)]
+pub struct ClusterConfig {
+    pub(crate) connection_timeout: Option<Duration>,
+    pub(crate) response_timeout: Option<Duration>,
+    #[cfg(feature = "cluster-async")]
+    pub(crate) async_push_sender: Option<std::sync::Arc<dyn crate::aio::AsyncPushSender>>,
+    #[cfg(feature = "cluster-async")]
+    pub(crate) async_dns_resolver: Option<std::sync::Arc<dyn crate::io::AsyncDNSResolver>>,
+}
+
+impl ClusterConfig {
+    /// Creates a new instance of the options with nothing set
+    pub fn new() -> Self {
+        Self::default()
+    }
+
+    /// Sets the connection timeout
+    pub fn set_connection_timeout(mut self, connection_timeout: std::time::Duration) -> Self {
+        self.connection_timeout = Some(connection_timeout);
+        self
+    }
+
+    /// Sets the response timeout
+    pub fn set_response_timeout(mut self, response_timeout: std::time::Duration) -> Self {
+        self.response_timeout = Some(response_timeout);
+        self
+    }
+
+    #[cfg(feature = "cluster-async")]
+    /// Sets a sender to receive pushed values.
+    ///
+    /// The sender can be a channel, or an arbitrary function that handles [crate::PushInfo] values.
+    /// This will fail client creation if the connection isn't configured for RESP3 communications via the [crate::RedisConnectionInfo::protocol] field.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// # use redis::cluster::ClusterConfig;
+    /// let (tx, mut rx) = tokio::sync::mpsc::unbounded_channel();
+    /// let config = ClusterConfig::new().set_push_sender(tx);
+    /// ```
+    ///
+    /// ```rust
+    /// # use std::sync::{Mutex, Arc};
+    /// # use redis::cluster::ClusterConfig;
+    /// let messages = Arc::new(Mutex::new(Vec::new()));
+    /// let config = ClusterConfig::new().set_push_sender(move |msg|{
+    ///     let Ok(mut messages) = messages.lock() else {
+    ///         return Err(redis::aio::SendError);
+    ///     };
+    ///     messages.push(msg);
+    ///     Ok(())
+    /// });
+    pub fn set_push_sender(mut self, sender: impl crate::aio::AsyncPushSender) -> Self {
+        self.async_push_sender = Some(std::sync::Arc::new(sender));
+        self
+    }
+
+    /// Set asynchronous DNS resolver for the underlying TCP connection.
+    ///
+    /// The parameter resolver must implement the [`crate::io::AsyncDNSResolver`] trait.
+    #[cfg(feature = "cluster-async")]
+    pub fn set_dns_resolver(mut self, resolver: impl crate::io::AsyncDNSResolver) -> Self {
+        self.async_dns_resolver = Some(std::sync::Arc::new(resolver));
+        self
+    }
+}
+
+/// This represents a Redis Cluster connection.
+///
+/// It stores the underlying connections maintained for each node in the cluster,
+/// as well as common parameters for connecting to nodes and executing commands.
 pub struct ClusterConnection<C = Connection> {
     initial_nodes: Vec<ConnectionInfo>,
     connections: RefCell<HashMap<String, C>>,
@@ -238,7 +319,7 @@ where
             connections: RefCell::new(HashMap::new()),
             slots: RefCell::new(SlotMap::new(cluster_params.read_from_replicas)),
             auto_reconnect: RefCell::new(true),
-            read_timeout: RefCell::new(Some(cluster_params.response_timeout)),
+            read_timeout: RefCell::new(cluster_params.response_timeout),
             write_timeout: RefCell::new(None),
             initial_nodes: initial_nodes.to_vec(),
             cluster_params,
@@ -320,22 +401,54 @@ where
     /// `BrokenPipe` error.
     fn create_initial_connections(&self) -> RedisResult<()> {
         let mut connections = HashMap::with_capacity(self.initial_nodes.len());
+        let mut failed_connections = Vec::new();
 
         for info in self.initial_nodes.iter() {
             let addr = info.addr.to_string();
 
-            if let Ok(mut conn) = self.connect(&addr) {
-                if conn.check_connection() {
-                    connections.insert(addr, conn);
-                    break;
+            match self.connect(&addr) {
+                Ok(mut conn) => {
+                    if conn.check_connection() {
+                        connections.insert(addr, conn);
+                        break;
+                    } else {
+                        failed_connections.push((
+                            addr,
+                            RedisError::from((
+                                ErrorKind::IoError,
+                                "Node failed to respond to connection check,",
+                            )),
+                        ));
+                    }
+                }
+                Err(conn_err) => {
+                    failed_connections.push((addr, conn_err));
                 }
             }
         }
 
         if connections.is_empty() {
+            // Create a composite description of why connecting to each node failed.
+            let detail = if failed_connections.is_empty() {
+                "List of initial nodes is empty".to_string()
+            } else {
+                let mut formatted_detail = "Failed to connect to each cluster node (".to_string();
+
+                for (index, (addr, conn_err)) in failed_connections.into_iter().enumerate() {
+                    if index != 0 {
+                        formatted_detail += "; ";
+                    }
+                    use std::fmt::Write;
+                    let _ = write!(&mut formatted_detail, "{addr}: {conn_err}");
+                }
+                formatted_detail += ")";
+                formatted_detail
+            };
+
             return Err(RedisError::from((
                 ErrorKind::IoError,
                 "It failed to check startup nodes.",
+                detail,
             )));
         }
 
@@ -424,17 +537,17 @@ where
         &self,
         connections: &'a mut HashMap<String, C>,
         route: &Route,
-    ) -> RedisResult<(String, &'a mut C)> {
+    ) -> (String, RedisResult<&'a mut C>) {
         let slots = self.slots.borrow();
         if let Some(addr) = slots.slot_addr_for_route(route) {
-            Ok((
+            (
                 addr.to_string(),
-                self.get_connection_by_addr(connections, addr)?,
-            ))
+                self.get_connection_by_addr(connections, addr),
+            )
         } else {
             // try a random node next.  This is safe if slots are involved
             // as a wrong node would reject the request.
-            Ok(get_random_connection(connections))
+            get_random_connection_or_error(connections)
         }
     }
 
@@ -465,9 +578,9 @@ where
 
         match RoutingInfo::for_routable(cmd) {
             Some(RoutingInfo::SingleNode(SingleNodeRoutingInfo::Random)) => {
-                let mut rng = thread_rng();
+                let mut rng = rng();
                 Ok(addr_for_slot(Route::new(
-                    rng.gen_range(0..SLOT_SIZE),
+                    rng.random_range(0..SLOT_SIZE),
                     SlotAddr::Master,
                 ))?)
             }
@@ -501,31 +614,23 @@ where
         &'a self,
         input: Input,
         addresses: HashSet<&'a str>,
-        connections: &'a mut HashMap<String, C>,
     ) -> Vec<RedisResult<(&'a str, Value)>> {
         addresses
             .into_iter()
             .map(|addr| {
-                let connection = self.get_connection_by_addr(connections, addr)?;
-                match input {
-                    Input::Slice { cmd, routable: _ } => connection
-                        .req_packed_command(cmd)
-                        .and_then(|value| value.extract_error()),
-                    Input::Cmd(cmd) => connection
-                        .req_command(cmd)
-                        .and_then(|value| value.extract_error()),
-                    Input::Commands {
-                        cmd: _,
-                        route: _,
-                        offset: _,
-                        count: _,
-                    } => Err((
-                        ErrorKind::ClientError,
-                        "req_packed_commands isn't supported with multiple nodes",
-                    )
-                        .into()),
-                }
-                .map(|res| (addr, res))
+                let (host, port) = split_node_address(addr).unwrap();
+                self.request(
+                    input.clone(),
+                    Some(RoutingInfo::SingleNode(SingleNodeRoutingInfo::ByAddress {
+                        host: host.to_string(),
+                        port,
+                    })),
+                )
+                .map(|res| match res {
+                    Output::Single(value) => (addr, value),
+                    // technically this shouldn't be possible, but I prefer not to crash here.
+                    Output::Multi(values) => (addr, Value::Array(values)),
+                })
             })
             .collect()
     }
@@ -534,18 +639,16 @@ where
         &'a self,
         input: Input,
         slots: &'a mut SlotMap,
-        connections: &'a mut HashMap<String, C>,
     ) -> Vec<RedisResult<(&'a str, Value)>> {
-        self.execute_on_all(input, slots.addresses_for_all_nodes(), connections)
+        self.execute_on_all(input, slots.addresses_for_all_nodes())
     }
 
     fn execute_on_all_primaries<'a>(
         &'a self,
         input: Input,
         slots: &'a mut SlotMap,
-        connections: &'a mut HashMap<String, C>,
     ) -> Vec<RedisResult<(&'a str, Value)>> {
-        self.execute_on_all(input, slots.addresses_for_all_primaries(), connections)
+        self.execute_on_all(input, slots.addresses_for_all_primaries())
     }
 
     fn execute_multi_slot<'a, 'b>(
@@ -589,20 +692,31 @@ where
                 self.execute_multi_slot(input, &mut slots, &mut connections, routes)
             }
             MultipleNodeRoutingInfo::AllMasters => {
-                self.execute_on_all_primaries(input, &mut slots, &mut connections)
+                drop(connections);
+                self.execute_on_all_primaries(input, &mut slots)
             }
             MultipleNodeRoutingInfo::AllNodes => {
-                self.execute_on_all_nodes(input, &mut slots, &mut connections)
+                drop(connections);
+                self.execute_on_all_nodes(input, &mut slots)
             }
         };
 
         match response_policy {
             Some(ResponsePolicy::AllSucceeded) => {
+                let mut last_result = None;
                 for result in results {
-                    result?;
+                    last_result = Some(result?);
                 }
 
-                Ok(Value::Okay)
+                last_result
+                    .ok_or(
+                        (
+                            ErrorKind::ClusterConnectionNotFound,
+                            "No results received for multi-node operation",
+                        )
+                            .into(),
+                    )
+                    .map(|(_, res)| res)
             }
             Some(ResponsePolicy::OneSucceeded) => {
                 let mut last_failure = None;
@@ -674,17 +788,7 @@ where
     }
 
     #[allow(clippy::unnecessary_unwrap)]
-    fn request(&self, input: Input) -> RedisResult<Output> {
-        let route_option = match &input {
-            Input::Slice { cmd: _, routable } => RoutingInfo::for_routable(routable),
-            Input::Cmd(cmd) => RoutingInfo::for_routable(*cmd),
-            Input::Commands {
-                cmd: _,
-                route,
-                offset: _,
-                count: _,
-            } => Some(RoutingInfo::SingleNode(route.clone())),
-        };
+    fn request(&self, input: Input, route_option: Option<RoutingInfo>) -> RedisResult<Output> {
         let single_node_routing = match route_option {
             Some(RoutingInfo::SingleNode(single_node_routing)) => single_node_routing,
             Some(RoutingInfo::MultiNode((multi_node_routing, response_policy))) => {
@@ -707,34 +811,53 @@ where
                         Redirect::Moved(addr) => (addr, false),
                         Redirect::Ask(addr) => (addr, true),
                     };
-                    let conn = self.get_connection_by_addr(&mut connections, &addr)?;
+                    let mut conn = self.get_connection_by_addr(&mut connections, &addr);
                     if is_asking {
                         // if we are in asking mode we want to feed a single
                         // ASKING command into the connection before what we
                         // actually want to execute.
-                        conn.req_packed_command(&b"*1\r\n$6\r\nASKING\r\n"[..])
-                            .and_then(|value| value.extract_error())?;
+                        conn = conn.and_then(|conn| {
+                            conn.req_packed_command(&b"*1\r\n$6\r\nASKING\r\n"[..])
+                                .and_then(|value| value.extract_error())?;
+                            Ok(conn)
+                        });
                     }
-                    (addr.to_string(), conn)
+                    (addr, conn)
                 } else {
                     match &single_node_routing {
-                        SingleNodeRoutingInfo::Random => get_random_connection(&mut connections),
+                        SingleNodeRoutingInfo::Random => {
+                            get_random_connection_or_error(&mut connections)
+                        }
                         SingleNodeRoutingInfo::SpecificNode(route) => {
-                            self.get_connection(&mut connections, route)?
+                            self.get_connection(&mut connections, route)
                         }
                         SingleNodeRoutingInfo::ByAddress { host, port } => {
                             let address = format!("{host}:{port}");
-                            let conn = self.get_connection_by_addr(&mut connections, &address)?;
+                            let conn = self.get_connection_by_addr(&mut connections, &address);
                             (address, conn)
                         }
                     }
                 };
-                (addr, input.send(conn))
+                (addr, conn.and_then(|conn| input.send(conn)))
             };
 
             match rv {
                 Ok(rv) => return Ok(rv),
                 Err(err) => {
+                    if err.kind() == ErrorKind::ClusterConnectionNotFound
+                        && *self.auto_reconnect.borrow()
+                    {
+                        for node in &self.initial_nodes {
+                            let addr = node.addr.to_string();
+                            if let Ok(mut conn) = self.connect(&addr) {
+                                if conn.check_connection() {
+                                    self.connections.borrow_mut().insert(addr, conn);
+                                }
+                            }
+                        }
+                        self.refresh_slots()?;
+                    }
+
                     if retries == self.cluster_params.retry_params.number_of_retries {
                         return Err(err);
                     }
@@ -764,6 +887,8 @@ where
                         }
                         crate::types::RetryMethod::Reconnect => {
                             if *self.auto_reconnect.borrow() {
+                                // if the connection is no longer valid, we should remove it.
+                                self.connections.borrow_mut().remove(&addr);
                                 if let Ok(mut conn) = self.connect(&addr) {
                                     if conn.check_connection() {
                                         self.connections.borrow_mut().insert(addr, conn);
@@ -813,7 +938,8 @@ where
         // retry logic that handles these cases.
         for retry_idx in to_retry {
             let cmd = &cmds[retry_idx];
-            results[retry_idx] = self.request(Input::Cmd(cmd))?.into();
+            let routing = RoutingInfo::for_routable(cmd);
+            results[retry_idx] = self.request(Input::Cmd(cmd), routing)?.into();
         }
         Ok(results)
     }
@@ -857,6 +983,12 @@ where
             None => Ok(to_retry),
         }
     }
+
+    /// Send a command to the given `routing`.
+    pub fn route_command(&mut self, cmd: &Cmd, routing: RoutingInfo) -> RedisResult<Value> {
+        self.request(Input::Cmd(cmd), Some(routing))
+            .map(|res| res.into())
+    }
 }
 
 const MULTI: &[u8] = "*1\r\n$5\r\nMULTI\r\n".as_bytes();
@@ -866,7 +998,8 @@ impl<C: Connect + ConnectionLike> ConnectionLike for ClusterConnection<C> {
     }
 
     fn req_command(&mut self, cmd: &Cmd) -> RedisResult<Value> {
-        self.request(Input::Cmd(cmd)).map(|res| res.into())
+        let routing = RoutingInfo::for_routable(cmd);
+        self.request(Input::Cmd(cmd), routing).map(|res| res.into())
     }
 
     fn req_packed_command(&mut self, cmd: &[u8]) -> RedisResult<Value> {
@@ -876,10 +1009,14 @@ impl<C: Connect + ConnectionLike> ConnectionLike for ClusterConnection<C> {
             cmd
         };
         let value = parse_redis_value(actual_cmd)?;
-        self.request(Input::Slice {
-            cmd,
-            routable: value,
-        })
+        let routing = RoutingInfo::for_routable(&value);
+        self.request(
+            Input::Slice {
+                cmd,
+                routable: value,
+            },
+            routing,
+        )
         .map(|res| res.into())
     }
 
@@ -896,17 +1033,16 @@ impl<C: Connect + ConnectionLike> ConnectionLike for ClusterConnection<C> {
         };
         let value = parse_redis_value(actual_cmd)?;
         let route = match RoutingInfo::for_routable(&value) {
+            // we don't allow routing multiple commands to multiple nodes.
             Some(RoutingInfo::MultiNode(_)) => None,
             Some(RoutingInfo::SingleNode(route)) => Some(route),
             None => None,
         }
         .unwrap_or(SingleNodeRoutingInfo::Random);
-        self.request(Input::Commands {
-            cmd,
-            offset,
-            count,
-            route,
-        })
+        self.request(
+            Input::Commands { cmd, offset, count },
+            Some(RoutingInfo::SingleNode(route)),
+        )
         .map(|res| res.into())
     }
 
@@ -953,18 +1089,28 @@ impl NodeCmd {
     }
 }
 
-// TODO: This function can panic and should probably
-// return an Option instead:
 fn get_random_connection<C: ConnectionLike + Connect + Sized>(
     connections: &mut HashMap<String, C>,
-) -> (String, &mut C) {
-    let addr = connections
-        .keys()
-        .choose(&mut thread_rng())
-        .expect("Connections is empty")
-        .to_string();
-    let con = connections.get_mut(&addr).expect("Connections is empty");
-    (addr, con)
+) -> Option<(String, &mut C)> {
+    let addr = connections.keys().choose(&mut rng())?.to_string();
+    let conn = connections.get_mut(&addr)?;
+    Some((addr, conn))
+}
+
+fn get_random_connection_or_error<C: ConnectionLike + Connect + Sized>(
+    connections: &mut HashMap<String, C>,
+) -> (String, RedisResult<&mut C>) {
+    match get_random_connection(connections) {
+        Some((addr, conn)) => (addr, Ok(conn)),
+        None => (
+            // we need to add a fake address in order for the error to be handled - the code that uses it assumes there's an address attached.
+            String::new(),
+            Err(RedisError::from((
+                ErrorKind::ClusterConnectionNotFound,
+                "No connections found",
+            ))),
+        ),
+    }
 }
 
 // The node string passed to this function will always be in the format host:port as it is either:
@@ -974,24 +1120,10 @@ pub(crate) fn get_connection_info(
     node: &str,
     cluster_params: ClusterParams,
 ) -> RedisResult<ConnectionInfo> {
-    let invalid_error = || (ErrorKind::InvalidClientConfig, "Invalid node string");
-
-    let (host, port) = node
-        .rsplit_once(':')
-        .and_then(|(host, port)| {
-            Some(host.trim_start_matches('[').trim_end_matches(']'))
-                .filter(|h| !h.is_empty())
-                .zip(u16::from_str(port).ok())
-        })
-        .ok_or_else(invalid_error)?;
+    let (host, port) = split_node_address(node)?;
 
     Ok(ConnectionInfo {
-        addr: get_connection_addr(
-            host.to_string(),
-            port,
-            cluster_params.tls,
-            cluster_params.tls_params,
-        ),
+        addr: get_connection_addr(host, port, cluster_params.tls, cluster_params.tls_params),
         redis: RedisConnectionInfo {
             password: cluster_params.password,
             username: cluster_params.username,
@@ -1001,6 +1133,19 @@ pub(crate) fn get_connection_info(
     })
 }
 
+fn split_node_address(node: &str) -> RedisResult<(String, u16)> {
+    let invalid_error =
+        || RedisError::from((ErrorKind::InvalidClientConfig, "Invalid node string"));
+    node.rsplit_once(':')
+        .and_then(|(host, port)| {
+            Some(host.trim_start_matches('[').trim_end_matches(']'))
+                .filter(|h| !h.is_empty())
+                .map(str::to_string)
+                .zip(u16::from_str(port).ok())
+        })
+        .ok_or_else(invalid_error)
+}
+
 pub(crate) fn get_connection_addr(
     host: String,
     port: u16,
